package com.sleepycat.je.dbi;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import com.sleepycat.je.DatabaseException;
import com.sleepycat.je.txn.BasicLocker;
import com.sleepycat.je.txn.Locker;
import com.sleepycat.je.txn.Txn;

public privileged aspect TxnDeleteSupport {
	/*
	 * We have to keep a set of DatabaseCleanupInfo objects so after commit or
	 * abort of Environment.truncateDatabase() or Environment.removeDatabase(),
	 * we can appropriately purge the unneeded MapLN and DatabaseImpl.
	 * Synchronize access to this set on this object.
	 */
	private Set Txn.deletedDatabases;

	
	/**
	 * @param dbImpl
	 *            databaseImpl to remove
	 * @param deleteAtCommit
	 *            true if this databaseImpl should be cleaned on commit, false
	 *            if it should be cleaned on abort.
	 * @param mb
	 *            environment memory budget.
	 */
	public void Txn.markDeleteAtTxnEnd(DatabaseImpl dbImpl, boolean deleteAtCommit)
			throws DatabaseException {

		synchronized (this) {
			int delta = 0;
			if (deletedDatabases == null) {
				deletedDatabases = new HashSet();
				delta += MemoryBudget.HASHSET_OVERHEAD;
			}

			deletedDatabases
					.add(new DatabaseCleanupInfo(dbImpl, deleteAtCommit));
			delta += MemoryBudget.HASHSET_ENTRY_OVERHEAD
					+ MemoryBudget.OBJECT_OVERHEAD;
			updateMemoryUsage(delta);
		}
	}


	/**
	 * Database operations like remove and truncate leave behind residual
	 * DatabaseImpls that must be purged at transaction commit or abort.
	 */
	public abstract void Locker.markDeleteAtTxnEnd(DatabaseImpl db,
			boolean deleteAtCommit) throws DatabaseException;
	
	public void BasicLocker.markDeleteAtTxnEnd(DatabaseImpl db,
			boolean deleteAtCommit) throws DatabaseException {

		if (deleteAtCommit) {
			db.deleteAndReleaseINs();
		}
	}
	
	
	
	/*
	 * Leftover databaseImpls that are a by-product of database operations like
	 * removeDatabase(), truncateDatabase() will be deleted after the write
	 * locks are released. However, do set the database state appropriately
	 * before the locks are released.
	 */
	private void Txn.setDeletedDatabaseState(boolean isCommit)
			throws DatabaseException {

		if (deletedDatabases != null) {
			Iterator iter = deletedDatabases.iterator();
			while (iter.hasNext()) {
				DatabaseCleanupInfo info = (DatabaseCleanupInfo) iter.next();
				if (info.deleteAtCommit == isCommit) {
					info.dbImpl.startDeleteProcessing();
				}
			}
		}
	}
	
	/**
	 * Cleanup leftover databaseImpls that are a by-product of database
	 * operations like removeDatabase(), truncateDatabase().
	 * 
	 * This method must be called outside the synchronization on this txn,
	 * because it calls deleteAndReleaseINs, which gets the TxnManager's allTxns
	 * latch. The checkpointer also gets the allTxns latch, and within that
	 * latch, needs to synchronize on individual txns, so we must avoid a
	 * latching hiearchy conflict.
	 */
	private void Txn.cleanupDatabaseImpls(boolean isCommit)
			throws DatabaseException {

		if (deletedDatabases != null) {
			/* Make a copy of the deleted databases while synchronized. */
			DatabaseCleanupInfo[] infoArray;
			synchronized (this) {
				infoArray = new DatabaseCleanupInfo[deletedDatabases.size()];
				deletedDatabases.toArray(infoArray);
			}
			for (int i = 0; i < infoArray.length; i += 1) {
				DatabaseCleanupInfo info = infoArray[i];
				if (info.deleteAtCommit == isCommit) {
					info.dbImpl.releaseDeletedINs();
				}
			}
			deletedDatabases = null;
		}
	}
	
	
	
	after(Txn txn) throws DatabaseException:
		call(void Txn.hook_commitDeletedDatabaseState()) && this(txn){
		/*
		 * Set database state for deletes before releasing any write
		 * locks.
		 */
		txn.setDeletedDatabaseState(true);
	}
	
	after(Txn txn) throws DatabaseException:
		call(void Txn.hook_abortDeletedDatabaseState()) && this(txn){
		/*
		 * Set database state for deletes before releasing any write
		 * locks.
		 */
		txn.setDeletedDatabaseState(false);

	}
	
	before(Txn txn) throws DatabaseException:
		call(void Txn.close(boolean)) && withincode(long Txn.commit(..)) && this(txn){
		/*
		 * Purge any databaseImpls not needed as a result of the commit. Be
		 * sure to do this outside the synchronization block, to avoid
		 * conflict w/checkpointer.
		 */
			txn.cleanupDatabaseImpls(true);
	}
	
	before(Txn txn) throws DatabaseException:
		call(void Txn.hook_cleanupDatabaseImpls()) && withincode(* Txn.abortInternal(..)) && this(txn){
		/*
		 * Purge any databaseImpls not needed as a result of the abort. Be
		 * sure to do this outside the synchronization block, to avoid
		 * conflict w/checkpointer.
		 */
		txn.cleanupDatabaseImpls(false);

	}
}
